local AddonName, a = ...
if a.BuildFail(50000) then return end
local s = SpellFlashAddon
local x = s.UpdatedVariables
local c = BittensSpellFlashLibrary

local UnitLevel = UnitLevel
local GetPowerRegen = GetPowerRegen
local GetSpellBookItemName = GetSpellBookItemName
local GetTime = GetTime
local IsSwimming = IsSwimming
local SPELL_POWER_BURNING_EMBERS = SPELL_POWER_BURNING_EMBERS
local SPELL_POWER_SOUL_SHARDS = SPELL_POWER_SOUL_SHARDS
local UnitExists = UnitExists
local math = math
local select = select

c.Init(a)

------------------------------------------------------------------------ Common
local petIdentifiers = {
	[s.SpellName(c.GetID("Firebolt"), true)] = "Imp",
	[s.SpellName(c.GetID("Felbolt"), true)] = "Fel Imp",
	[s.SpellName(c.GetID("Torment"), true)] = "Voidwalker",
	[s.SpellName(c.GetID("Tongue Lash"), true)] = "Observer",
	[s.SpellName(c.GetID("Shadow Bite"), true)] = "Felhunter",
	[s.SpellName(c.GetID("Lash of Pain"), true)] = "Succubus",
}

function a.GetCurrentPet()
	for i = 1, 10000 do
		local spell = GetSpellBookItemName(i, "pet")
		if spell == nil then
			return nil
		end
		
		local pet = petIdentifiers[spell]
		if pet ~= nil then
			return pet
		end
	end
end

c.AddOptionalSpell("Dark Intent", nil, {
	Override = function()
		return c.RaidBuffNeeded(c.SPELL_POWER_BUFFS)
	end
})

c.AddOptionalSpell("Curse of the Elements", nil, {
	Debuff = c.MAGIC_VULNERABILITY_DEBUFFS
})

c.AddOptionalSpell("Soulstone", nil, {
	NoRangeCheck = true,
	CheckFirst = function()
		return not s.InRaidOrParty() and c.SelfBuffNeeded("Soulstone")
	end
})

c.AddOptionalSpell("Dark Regeneration", nil, {
	CheckFirst = function(z)
		if s.HealthPercent("player") < 70 then
			z.FlashSize = nil
			return true
		end
		
		if UnitExists("pet") and s.HealthPercent("pet") < 70 then
			z.FlashSize = s.FlashSizePercent() / 2
			return true
		end
	end
})

c.AddOptionalSpell("Underwater Breath", nil, {
	Override = function()
		return IsSwimming() and c.SelfBuffNeeded("Underwater Breath")
	end
})

c.AddInterrupt("Spell Lock", nil, {
	NoGCD = true,
})

c.AddInterrupt("Command Spell Lock", nil, {
	NoGCD = true,
})

c.AddInterrupt("Optical Blast", nil, {
	NoGCD = true,
})

-------------------------------------------------------------------- Affliction
local baseLengths = {
	["Agony"] = 24,
	["Corruption"] = 18,
	["Unstable Affliction"] = 14,
}

local function expiresWithin(name, time)
	return not c.IsCastingOrInAir(name) 
		and not c.IsQueued("Soul Swap") 
		and c.GetMyDebuffDuration(name) < time
		and (name ~= "Corruption" 
			or (not a.SoCEmpowered and GetTime() - a.SoCExplodeTime > 1))
end

local function gcdCheck(name, padding)
	local time = math.max(1, c.GetHastedTime(1.5))
	if padding ~= nil then
		time = time + padding
	end
	return expiresWithin(name, time)
end

local function soonCheck(name, considerCastTime)
	local time = a.Snapshots[name].Tick
	if s.HasSpell(c.GetID("Pandemic")) then
		time = 2 * time
	end
	if considerCastTime then
		time = time + c.GetCastTime(name)
	end
	return expiresWithin(name, time)
end

local function pandemicCheck(name, z, considerCastTime)
	if not s.HasSpell(c.GetID("Pandemic")) then
		return false
	end
	
	local time = baseLengths[name] / 2 - 1
	if considerCastTime then
		time = time + c.GetCastTime(name)
	end
	if not expiresWithin(name, time) then
		return false
	end
	
	if z ~= nil then
		local now = a.GetDps(name)
		local snapped = a.Snapshots[name].Dps
		if now < snapped then
			z.FlashColor = "red"
		elseif now == snapped then
			z.FlashColor = "yellow"
		else
			z.FlashColor = "green"
		end
	end
	return true
end

local function sizeForDarkSoul(z, ...)
	if a.Shards == 0 then
		z.FlashSize = nil
		return
	end
	
	for i = 1, select("#", ...) do
		local name = select(i, ...)
		if c.GetMyDebuffDuration(name) < c.GetCooldown("Dark Soul: Misery") 
			or c.HasBuff("Dark Soul: Misery") then
			
			z.FlashSize = nil
			return
		end
	end
	z.FlashSize = s.FlashSizePercent() / 2
end

c.RegisterForFullChannels("Malefic Grasp", 4)
c.AssociateTravelTimes(.8, "Haunt")

c.AddSpell("Soulburn", "under Dark Soul: Misery", {
	CheckFirst = function()
		if not c.HasBuff("Dark Soul: Misery") 
			or c.HasBuff("Soulburn") 
			or c.IsCasting("Soulburn") then
			
			return false
		elseif s.HasSpell(c.GetID("Pandemic")) then
			return pandemicCheck("Agony")
				or pandemicCheck("Unstable Affliction")
				or pandemicCheck("Corruption")
		else
			return soonCheck("Agony")
				or soonCheck("Unstable Affliction")
				or soonCheck("Corruption")
		end
	end
})

c.AddSpell("Soulburn", "during Execute", {
	CheckFirst = function(z)
		if c.HasBuff("Soulburn") or c.IsCasting("Soulburn") then
			return false
		end
		
		sizeForDarkSoul(z, "Agony", "Corruption", "Unstable Affliction")
		
		if gcdCheck("Agony", .3)
			or gcdCheck("Unstable Affliction")
			or gcdCheck("Corruption") then
			
			z.FlashColor = nil
			z.Continue = nil
			return true
		end
		
		if pandemicCheck("Agony", z)
			or pandemicCheck("Unstable Affliction", z)
			or pandemicCheck("Corruption", z) then
			
			if c.HasBuff("Dark Soul: Misery") then
				z.FlashColor = nil
				z.Continue = nil
			else
				z.Continue = true
			end
			return true
		end
	end
})

c.AddSpell("Soul Swap", "under Soulburn", {
	CheckFirst = function()
		return c.HasBuff("Soulburn")
	end
})

c.AddSpell("Soul Swap", "during Execute", {
	CheckFirst = function()
		return c.HasBuff("Soulburn")
	end
})

c.AddOptionalSpell("Dark Soul: Misery", nil, {
	Override = function()
		return a.Shards > 0 and c.GetCooldown("Dark Soul: Misery") == 0
	end
})

c.AddOptionalSpell("Grimoire: Felhunter")

c.AddSpell("Haunt", nil, {
	CheckFirst = function(z)
		if a.Shards == 0
			or not c.ShouldCastToRefresh("Haunt", "Haunt", 0, true) then
			
			return false
		end
		
		if a.Shards >= s.MaxPower("player", SPELL_POWER_SOUL_SHARDS) - 1 
			or c.HasBuff("Dark Soul: Misery") then
			
			z.FlashColor = nil
			z.Continue = nil
			z.FlashSize = nil
			return true
		end
		
		z.Continue = true
		z.FlashColor = "yellow"
		local miseryCD = c.GetCooldown("Dark Soul: Misery")
		if miseryCD < 35 then
			z.FlashSize = s.FlashSizePercent() / 2
		else
			z.FlashSize = nil
		end
		return true
	end
})

c.AddSpell("Haunt", "during Execute", {
	CheckFirst = function(z)
		return a.Shards > 0
			and c.ShouldCastToRefresh("Haunt", "Haunt", 0, true)
	end
})

c.AddSpell("Agony", "within GCD", {
	CheckFirst = function(z)
		sizeForDarkSoul(z, "Unstable Affliction")
		return gcdCheck("Agony", .3)
	end
})

c.AddOptionalSpell("Agony", "with Pandemic", {
	CheckFirst = function(z)
		sizeForDarkSoul(z, "Agony")
		return pandemicCheck("Agony", z)
	end
})

c.AddSpell("Agony", "Soon", {
	CheckFirst = function(z)
		sizeForDarkSoul(z, "Agony")
		return soonCheck("Agony")
	end
})

c.AddSpell("Agony", "before Malefic Grasp", {
	CheckFirst = function(z)
		sizeForDarkSoul(z, "Agony")
		return expiresWithin("Agony", c.GetHastedTime(4))
	end
})

c.AddSpell("Corruption", "within GCD", {
	CheckFirst = function(z)
		sizeForDarkSoul(z, "Corruption")
		return gcdCheck("Corruption")
	end
})

c.AddOptionalSpell("Corruption", "with Pandemic", {
	CheckFirst = function(z)
		sizeForDarkSoul(z, "Corruption")
		return pandemicCheck("Corruption", z)
	end
})

c.AddSpell("Corruption", "Soon", {
	CheckFirst = function(z)
		sizeForDarkSoul(z, "Corruption")
		return soonCheck("Corruption")
	end
})

c.AddSpell("Unstable Affliction", "within GCD", {
	CheckFirst = function(z)
		sizeForDarkSoul(z, "Unstable Affliction")
		return gcdCheck(
			"Unstable Affliction", c.GetCastTime("Unstable Affliction"))
	end
})

c.AddOptionalSpell("Unstable Affliction", "with Pandemic", {
	CheckFirst = function(z)
		sizeForDarkSoul(z, "Unstable Affliction")
		return pandemicCheck(
			"Unstable Affliction", 
			z,
			a.Shards == 0 or not c.HasBuff("Dark Soul: Misery"))
	end
})

c.AddSpell("Unstable Affliction", "Soon", {
	CheckFirst = function(z)
		sizeForDarkSoul(z, "Unstable Affliction")
		return soonCheck(
			"Unstable Affliction", 
			a.Shards == 0 or not c.HasBuff("Dark Soul: Misery"))
	end
})

c.AddOptionalSpell("Life Tap", "for Affliction", {
	CheckFirst = function()
		if c.HasBuff("Dark Soul: Misery") or c.HasBuff(c.BLOODLUST_BUFFS) then
			return false
		end
		
		if s.HealthPercent() <= 20 then
			return s.PowerPercent("player") < 10
		else
			return s.PowerPercent("player") < 50
		end
	end
})

-------------------------------------------------------------------- Demonology
c.AddOptionalSpell("Dark Soul: Knowledge")

c.AddOptionalSpell("Grimoire: Felguard")

c.AddOptionalSpell("Metamorphosis", nil, {
	Type = "form",
	CheckFirst = function()
		if a.Fury < 40 then
			return false
		end
		return a.Fury > 950
			or c.HasBuff("Dark Soul: Knowledge")
			or c.GetMyDebuffDuration("Corruption") < 5
	end
})

c.AddOptionalSpell("Metamorphosis", "Cancel", {
	Type = "form",
	Override = function(z)
		return a.Morphed
			and a.Fury < 650
			and not c.HasBuff("Dark Soul: Knowledge")
			and c.GetMyDebuffDuration("Corruption") > 20
	end
})

c.AddSpell("Corruption", "for Demonology", {
	FlashID = { "Corruption", "Doom" },
	CheckFirst = function()
		return not a.Morphed and not c.HasMyDebuff("Corruption")
	end
})

c.AddSpell("Doom", nil, {
	EarlyRefresh = 15,
	FlashID = { "Corruption", "Doom" },
	Override = function(z)
		return a.Morphed
			and a.Fury >= 60
			and not c.IsAuraPendingFor("Doom")
			and (c.GetMyDebuffDuration("Doom") < z.EarlyRefresh
				or (not a.DoomIsBuffed and c.HasBuff("Dark Soul: Knowledge")))
	end
})
c.ManageDotRefresh("Doom", 15)

c.AddSpell("Hand of Gul'dan", nil, {
	CheckFirst = function()
		c.Debug("hand",not a.Morphed,
			not c.HasMyDebuff("Shadowflame"),
			not c.IsAuraPendingFor("Hand of Gul'dan"))
		return not a.Morphed
			and not c.HasMyDebuff("Shadowflame")
			and not c.IsAuraPendingFor("Hand of Gul'dan")
	end
})
c.RegisterAura("Hand of Gul'dan", "Shadowflame")

c.AddSpell("Touch of Chaos", nil, {
	FlashID = { "Touch of Chaos", "Shadow Bolt" },
	Override = function()
		return a.Morphed and a.Fury >= 40
	end
})

c.AddSpell("Touch of Chaos", "for Corruption", {
	FlashID = { "Touch of Chaos", "Shadow Bolt" },
	Override = function()
		local duration = c.GetMyDebuffDuration("Corruption")
		return a.Morphed and a.Fury >= 40 and duration > 0 and duration < 20
	end
})

c.AddSpell("Soul Fire", nil, {
	CheckFirst = function()
		local stacks = c.GetBuffStack("Molten Core")
		if c.IsCasting("Soul Fire") then
			stacks = stacks - 1
		end
		return stacks > 0
			and (not a.Morphed or a.Fury >= 80)
	end
})

c.AddSpell("Shadow Bolt", nil, {
	FlashID = { "Touch of Chaos", "Shadow Bolt" },
	CheckFirst = function()
		return not a.Morphed
	end
})

c.AddOptionalSpell("Life Tap", "under 50", {
	CheckFirst = function()
		return s.PowerPercent("player") < 50
	end
})

c.AddInterrupt("Axe Toss", nil, {
	NoGCD = true,
})

c.AddInterrupt("Super Axe Toss", nil, {
	NoGCD = true,
})

------------------------------------------------------------------- Destruction
c.AddOptionalSpell("Dark Soul: Instability")

c.AddOptionalSpell("Grimoire: Imp")

c.AddSpell("Immolate", nil, {
	FlashID = { "Immolate", "Immolate AoE" },
	CheckFirst = function()
		if c.IsCastingOrInAir("Immolate") 
			or c.IsCastingOrInAir("Immolate AoE") then
			
			return false
		end
		
		local cinCast = c.GetCastTime("Incinerate")
		local realDraft = s.Buff(c.GetID("Backdraft"), "player")
		if a.Backdraft > 0 and not realDraft then
			cinCast = cinCast / 1.3
		elseif a.Backdraft == 0 and realDraft then
			cinCast = cinCast * 1.3
		end
		return math.max(
				c.GetMyDebuffDuration("Immolate"), 
				c.GetMyDebuffDuration("Immolate AoE")) 
			< c.GetCastTime("Immolate") + cinCast + .1
	end
})

c.AddSpell("Chaos Bolt", nil, {
	CheckFirst = function()
		return a.Backdraft < 3
			and (c.GetPower(select(2, GetPowerRegen())) / s.MaxPower("player") 
					< .8
				or not s.InRaidOrParty())
			and (s.MaxPower("player", SPELL_POWER_BURNING_EMBERS) - a.Embers 
					< .5
				or c.GetBuffDuration("Dark Soul: Instability") 
					> c.GetCastTime("Chaos Bolt"))
	end
})

c.AddSpell("Conflagrate", nil, {
	CheckFirst = function()
		return a.Backdraft == 0
			and (not c.HasBuff("Backlash") 
				or c.IsCasting("Incinerate") 
				or c.IsCasting("Incinerate AoE"))
	end
})

c.AddSpell("Incinerate", nil, {
	FlashID = { "Incinerate", "Incinerate AoE" },
})

c.AddOptionalSpell("Ember Tap", nil, {
	CheckFirst = function()
		return s.HealthPercent("player") < 75
	end
})

c.AddOptionalSpell("Rain of Fire", nil, {
	MyDebuff = "Rain of Fire",
})
